@@ -3,7 +3,7 @@
 ;; Copyright (C) 2016 Josef Vlach
 ;; License: http://www.gnu.org/licenses/gpl.html
 
-;; Homepage: https://github.com/ensime/emacs-sbt-mode
+;; Homepage: https://github.com/hvesalai/emacs-sbt-mode
 ;; Keywords: languages
 ;; Package-Requires: ()
 
@@ -40,6 +40,7 @@
 (defvar-local sbt-hydra:projects nil)               ;; dir-local
 (defvar-local sbt-hydra:command-line-arguments nil) ;; dir-local
 (defvar-local sbt-hydra:system-properties nil)      ;; dir-local
+(defvar sbt-hydra:after-create-hook nil "Hook run after Hydra is created.")
 
 ;; Make `sbt-hydra:projects' safe if its value is list of strings
 (put 'sbt-hydra:projects 'safe-local-variable
@@ -198,8 +199,8 @@ _q_ quit _o_ back _u_ testOnly _x_ clean -- -z %`sbt-hydra:sbt-test-substring
         ,(format "
 %s
 ^-^----------^-^--------------^-^-------^-^-----------^-^--------^-^--------^-^--------^-^-----------^-^-----
-_q_ quit     _o_ testHydra    _p_ parse _u_ testOnly  _a_ repeat _n_ no-op  _i_ edit   _s_ sbt-shell _e_ eof
-_c_ compile  _y_ test:compile _t_ test  _k_ testQuick _r_ run    _l_ clean  _d_ reload _v_ history   _h_ help
+_q_ quit     _o_ testHydra      _p_ parse _u_ testOnly  _a_ repeat _n_ no-op  _i_ edit   _s_ sbt-shell _e_ eof
+_c_ compile  _y_ Test / compile _t_ test  _k_ testQuick _r_ run    _l_ clean  _d_ reload _v_ history   _h_ help
 " switcher)
         ,@heads)))
 
@@ -210,13 +211,13 @@ _c_ compile  _y_ test:compile _t_ test  _k_ testQuick _r_ run    _l_ clean  _d_
   `((sbt-hydra:run-project-command "clean" ,project) nil))
 
 (defun sbt-hydra-command:test-test (project)
-  `((sbt-hydra:run-project-command "test:test" ,project) nil))
+  `((sbt-hydra:run-project-command "Test / test" ,project) nil))
 
 (defun sbt-hydra-command:test-test-quick (project)
-  `((sbt-hydra:run-project-command "test:testQuick" ,project) nil))
+  `((sbt-hydra:run-project-command "Test / testQuick" ,project) nil))
 
 (defun sbt-hydra-command:test-compile (project)
-  `((sbt-hydra:run-project-command "test:compile" ,project) nil))
+  `((sbt-hydra:run-project-command "Test / compile" ,project) nil))
 
 (defun sbt-hydra-command:compile (project)
   `((sbt-hydra:run-project-command "compile" ,project) nil))
@@ -253,7 +254,7 @@ _c_ compile  _y_ test:compile _t_ test  _k_ testQuick _r_ run    _l_ clean  _d_
   `((sbt-hydra:test-only-hydra-on) nil :exit t))
 
 (defun sbt-hydra-command:it-test (project)
-  `((sbt-hydra:run-project-command "it:test" ,project) nil))
+  `((sbt-hydra:run-project-command "It / test" ,project) nil))
 
 (defun sbt-hydra-command:quit ()
   `(nil nil))
@@ -287,33 +288,33 @@ Most action provided are scoped by active project. For example key 'c' will trig
 
 Available actions:
 
-A, B, C ...      - switch between project hydras, this part is generated dynamically on start by executing 'plugins' command
-                   in *sbt* buffer or it is populated by projects defined in `sbt-hydra:projects' Directory Local Variables
+A, B, C ...        - switch between project hydras, this part is generated dynamically on start by executing 'plugins' command
+                     in *sbt* buffer or it is populated by projects defined in `sbt-hydra:projects' Directory Local Variables
 
-q - quit         - close hydra
-c - compile      - execute 'compile' command for active project
-y - test:compile - execute 'test:compile' command for active project
-t - test         - execute 'test' command for active project
-k - testQuick    - execute 'testQuick' command for active project
-r - run          - execute 'run' command for active project
-l - clean        - execute 'clean' command for active project
-d - reload       - execute 'reload' command
-e - eof          - send end-of-file character to the sbt prompt, useful for example for terminating running play application
-i - edit         - edit last executed command and execute edited command
-a - repeat       - repeat last executed command again
-n - no-op        - execute 'name' command. This is convenient way how to prevent last command execute again. For example when
-                   you don't want to execute test command again since it takes long time to finish
-p - parse        - parse output of Sbt buffer for failed test. It must be run after 'test' command finish execution and there
-                   are failed tests. In this case at the end of output is list of classes containing some failed tests.
-                   This action will parse this classes names and it will create so called 'testHydra' which allow
-                   run these classes only one by one.
-o - testHydra    - switch to special hydra which is created by 'parse' action. More info in section 'Test Hydra'
-u - testOnly     - run 'testOnly' command for active project with substring parameter (-- -z) containing text from the line point
-                   is at. Works in Sbt buffer or Scala source file, but point must be at line contaning should text of the test.
-f - it:test      - execute 'it:test' command for active project
-v - history      - displays commands history for effective execution of old commands.
-                   When hydra is executed with one prefix argument C-u invoking history command allows to remove selected command.
-                   When hydra is executed with tow or more prefix arguments C-u invoking history command will clear the history.
+q - quit           - close hydra
+c - compile        - execute 'compile' command for active project
+y - Test / compile - execute 'Test / compile' command for active project
+t - test           - execute 'test' command for active project
+k - testQuick      - execute 'testQuick' command for active project
+r - run            - execute 'run' command for active project
+l - clean          - execute 'clean' command for active project
+d - reload         - execute 'reload' command
+e - eof            - send end-of-file character to the sbt prompt, useful for example for terminating running play application
+i - edit           - edit last executed command and execute edited command
+a - repeat         - repeat last executed command again
+n - no-op          - execute 'name' command. This is convenient way how to prevent last command execute again. For example when
+                     you don't want to execute test command again since it takes long time to finish
+p - parse          - parse output of Sbt buffer for failed test. It must be run after 'test' command finish execution and there
+                     are failed tests. In this case at the end of output is list of classes containing some failed tests.
+                     This action will parse this classes names and it will create so called 'testHydra' which allow
+                     run these classes only one by one.
+o - testHydra      - switch to special hydra which is created by 'parse' action. More info in section 'Test Hydra'
+u - testOnly       - run 'testOnly' command for active project with substring parameter (-- -z) containing text from the line point
+                     is at. Works in Sbt buffer or Scala source file, but point must be at line contaning should text of the test.
+f - It / test      - execute 'It / test' command for active project
+v - history        - displays commands history for effective execution of old commands.
+                     When hydra is executed with one prefix argument C-u invoking history command allows to remove selected command.
+                     When hydra is executed with tow or more prefix arguments C-u invoking history command will clear the history.
 
 *** Test Hydra ***
 
@@ -325,7 +326,7 @@ u - testOnly     - set substring (-- -z) for running one of defined test. By def
                    contains should text of the test
 x - clean        - reset substring (-- -z) to empty string
 
-[a] [b] [c] ...  - it run 'test:testOnly' for failed test optionally limited to (-- -z) substring. This section is generated
+[a] [b] [c] ...  - it run 'Test / testOnly' for failed test optionally limited to (-- -z) substring. This section is generated
                    dynamically by the 'parse' action from project hydra.
 ")
           (goto-char (point-min))
@@ -339,7 +340,7 @@ x - clean        - reset substring (-- -z) to empty string
         (is-revolver (member project sbt-hydra:revolver-projects)))
     (cond (is-jetty
            (sbt-hydra:run-run-project-command "jetty:start" project))
-          (is-revolver
+          ((and is-revolver (not is-play)) ;; Run as revolver only if it is not play project
            (sbt-hydra:run-run-project-command "reStart" project))
           ((or main-class is-play) ;; play projects and simple projects are run in the same way
            (sbt-hydra:run-run-project-command (format "run %s" (concat "" (cdr (assoc project sbt-hydra:command-line-arguments)))) project)
@@ -461,10 +462,10 @@ Read data from the file specified by `sbt-hydra:history-file'."
         (t
          (save-excursion
            (let ((project-folder (let* ((root (sbt:find-root))
-                                 (buffer-name (buffer-file-name (current-buffer))) ;; not every buffer represents a file
-                                 (path-from-root (when (and root buffer-name) (replace-regexp-in-string (concat ".*" (substring root 1)) "" buffer-name))))
-                            (when (and path-from-root (string-match "^\\([[:word:]]*\\)/" path-from-root))
-                              (match-string-no-properties 1 path-from-root))))
+                                        (buffer-name (buffer-file-name (current-buffer))) ;; not every buffer represents a file
+                                        (path-from-root (when (and root buffer-name) (replace-regexp-in-string (concat ".*" (substring root 1)) "" buffer-name))))
+                                   (when (and path-from-root (string-match "^\\([[:word:]]*\\)/" path-from-root))
+                                     (match-string-no-properties 1 path-from-root))))
                  (substring (let ((line (thing-at-point 'line)))
                               (when (string-match "\"\\(.*\\)\" \\(in\\|should\\)" line)
                                 (replace-regexp-in-string "\"" "" (match-string-no-properties 1 line)))))
@@ -521,8 +522,12 @@ Read data from the file specified by `sbt-hydra:history-file'."
    (if sbt-hydra:test-hydra-active
        (if (functionp 'sbt-test-hydra/body)
            (sbt-test-hydra/body)
-     	(sbt-hydra:no-test-hydra))
-     (funcall sbt-hydra:current-hydra))))
+     	 (sbt-hydra:no-test-hydra))
+     (if (functionp sbt-hydra:current-hydra)
+         (funcall sbt-hydra:current-hydra)
+       ;; If sbt-hydra:current-hydra is not a function here it means
+       ;; that parsing of sbt output failed. Let's try it again.
+       (sbt-hydra:create-hydra)))))
 
 (defun sbt-hydra:no-test-hydra ()
   (message "No hydra defined for testOnly command"))
@@ -642,25 +647,91 @@ Read data from the file specified by `sbt-hydra:history-file'."
                                (format "%s -- -z \"%s\"" command sbt-hydra:sbt-test-substring))))
     (sbt-hydra:run-project-command command-with-params project)))
 
+(defun sbt-hydra:project-info-by-version (sbt-output)
+  "Output of 'plugins' command changed with sbt 1.4. Pre sbt 1.4 output looks like this:
+
+In file:/Users/scala/example-project-1.3/
+	sbt.plugins.CorePlugin: enabled in core, web
+	sbt.ScriptedPlugin
+	sbt.plugins.SbtPlugin
+	sbt.plugins.SemanticdbPlugin: enabled in core, web
+>
+
+Starting with sbt 1.4 output looks like this:
+
+In build /Users/scala/example-project-1.4/:
+  Enabled plugins in core:
+    sbt.plugins.CorePlugin
+    sbt.plugins.SemanticdbPlugin
+  Enabled plugins in web:
+    sbt.plugins.CorePlugin
+    sbt.plugins.SemanticdbPlugin
+Plugins that are loaded to the build but not enabled in any subprojects:
+  sbt.ScriptedPlugin
+  sbt.plugins.SbtPlugin
+>
+"
+  (if (string-match "  Enabled plugins in \\(.*\\):$" sbt-output)
+      (sbt-hydra:projects-info sbt-output)
+    (sbt-hydra:projects-info-legacy sbt-output)))
+
+(defconst sbt-hydra:play-plugin-name "play.sbt.Play.*")
+(defconst sbt-hydra:jetty-plugin-name "com.earldouglas.xwp.JettyPlugin")
+(defconst sbt-hydra:revolver-plugin-name "spray.revolver.RevolverPlugin")
+
 (defun sbt-hydra:projects-info (sbt-output)
+  (let ((project-names)
+        (play-projects)
+        (jetty-projects)
+        (revolver-projects))
+    (dolist (project (sbt-hydra:split-to-projects sbt-output))
+      (let ((project-name (car project))
+            (project-plugins-str (cdr project)))
+        (push project-name project-names)
+        (when (string-match sbt-hydra:play-plugin-name project-plugins-str)
+          (push project-name play-projects))
+        (when (string-match sbt-hydra:jetty-plugin-name project-plugins-str)
+          (push project-name jetty-projects))
+        (when (string-match sbt-hydra:revolver-plugin-name project-plugins-str)
+          (push project-name revolver-projects))))
+    (sbt-hydra:setup-projects project-names play-projects jetty-projects revolver-projects)))
+
+(defun sbt-hydra:setup-projects (project-names play-projects jetty-projects revolver-projects)
+  (setq sbt-hydra:play-framework-projects play-projects)
+  (setq sbt-hydra:jetty-projects jetty-projects)
+  (setq sbt-hydra:revolver-projects revolver-projects)
+
+  (if sbt-hydra:projects
+      sbt-hydra:projects
+    project-names))
+
+(defun sbt-hydra:split-to-projects (sbt-output)
+  (let ((data)
+        (str (sbt-hydra:strip-not-enabled-plugins sbt-output)))
+    (dolist (split (split-string str "Enabled") data)
+      (when (string-match "^ plugins in \\(.*\\):$" split)
+        (push (cons (match-string 1 split) split) data)))))
+
+(defun sbt-hydra:strip-not-enabled-plugins (sbt-output)
+  (let ((position (string-match "Plugins that are loaded to the build but not enabled in any subprojects:" sbt-output)))
+    (if position
+        (substring-no-properties sbt-output 0 position)
+      sbt-output)))
+
+(defun sbt-hydra:projects-info-legacy (sbt-output)
   (let ((project-names (sbt-hydra:projects-for-plugin sbt-output "sbt.plugins.CorePlugin"))
-        (play-projects (sbt-hydra:projects-for-plugin sbt-output "play.sbt.Play"))
-        (jetty-projects (sbt-hydra:projects-for-plugin sbt-output "com.earldouglas.xwp.JettyPlugin"))
-        (revolver-projects (sbt-hydra:projects-for-plugin sbt-output "spray.revolver.RevolverPlugin")))
-
-    (setq sbt-hydra:play-framework-projects play-projects)
-    (setq sbt-hydra:jetty-projects jetty-projects)
-    (setq sbt-hydra:revolver-projects revolver-projects)
-
-    (if sbt-hydra:projects
-        sbt-hydra:projects
-      project-names)))
+        (play-projects (sbt-hydra:projects-for-plugin sbt-output sbt-hydra:play-plugin-name))
+        (jetty-projects (sbt-hydra:projects-for-plugin sbt-output sbt-hydra:jetty-plugin-name))
+        (revolver-projects (sbt-hydra:projects-for-plugin sbt-output sbt-hydra:revolver-plugin-name)))
+    (sbt-hydra:setup-projects project-names play-projects jetty-projects revolver-projects)))
 
 (defun sbt-hydra:projects-for-plugin (sbt-output plugin)
   (if (string-match (format "^[[:space:]]*%s: enabled in \\(.*\\)$" plugin) sbt-output)
       (split-string (match-string 1 sbt-output) "," t " ")
     nil))
 
+(defvar sbt-hydra:plugins-output nil) ;; Helper variable to accumulate output of "plugins" command.
+
 (defun sbt-hydra:create-hydra ()
   "Create hydras for current scala project. It will create one hydra for every sbt project.
 List of sbt projects is determined in two ways:
@@ -671,6 +742,7 @@ The easiest way to use second option is by running `add-dir-local-variable' comm
   (if (sbt:find-root)
       (progn
         (setq sbt-hydra:sbt-output-cleared "")
+        (setq sbt-hydra:plugins-output "")
         (let ((buffer-res (sbt-switch-to-active-sbt-buffer)))
           (if (or (bufferp buffer-res)
                   (equal buffer-res "Already in sbt buffer!"))
@@ -684,8 +756,7 @@ The easiest way to use second option is by running `add-dir-local-variable' comm
               (sbt:run-sbt)
               (sbt-switch-to-active-sbt-buffer)
               (hack-dir-local-variables-non-file-buffer)
-              (sbt:command "plugins")
-              (add-hook 'comint-output-filter-functions 'sbt-hydra:parse-plugins-info-skip-init)))))
+              (add-hook 'comint-output-filter-functions 'sbt-hydra:detect-when-launched)))))
     "Not in sbt project. Hydra can be generated only from sbt project. See `sbt:find-root' to get more informations."))
 
 (defun sbt-hydra:parse-main-class (sbt-output)
@@ -694,14 +765,12 @@ The easiest way to use second option is by running `add-dir-local-variable' comm
 (defun sbt-hydra:parse-plugins-info (sbt-output)
   (sbt-hydra:parse-sbt-output sbt-output 'sbt-hydra:generate-hydras-from-plugins-info 'sbt-hydra:parse-plugins-info))
 
-(defun sbt-hydra:parse-plugins-info-skip-init (sbt-output)
-  (sbt-hydra:parse-sbt-output-skip-init sbt-output))
-
-(defun sbt-hydra:parse-sbt-output-skip-init (sbt-output)
-  (let* ((output-cleared (replace-regexp-in-string ansi-color-regexp "" sbt-output)))
+(defun sbt-hydra:detect-when-launched (sbt-output)
+  (let ((output-cleared (replace-regexp-in-string ansi-color-regexp "" sbt-output)))
     (when (string-match sbt:sbt-prompt-regexp output-cleared)
-      (remove-hook 'comint-output-filter-functions 'sbt-hydra:parse-plugins-info-skip-init)
-      (add-hook 'comint-output-filter-functions 'sbt-hydra:parse-plugins-info))))
+      (remove-hook 'comint-output-filter-functions 'sbt-hydra:detect-when-launched)
+      (add-hook 'comint-output-filter-functions 'sbt-hydra:parse-plugins-info)
+      (sbt:command "plugins"))))
 
 (defun sbt-hydra:parse-sbt-output (sbt-output f hook)
   (let* ((output-cleared (replace-regexp-in-string ansi-color-regexp "" sbt-output)))
@@ -763,13 +832,17 @@ The easiest way to use second option is by running `add-dir-local-variable' comm
   (sbt-hydra:generate-hydras projects)
   (sbt-hydra:run-current-hydra)
   (sbt-hydra:load-history)
-  (message "Success hydra for projects %s created." projects))
+  (message "Success hydra for projects %s created." projects)
+  (run-hooks 'sbt-hydra:after-create-hook))
 
 (defun sbt-hydra:generate-hydras-from-plugins-info (sbt-output)
-  (sbt-hydra:generate-hydras-from-projects (sbt-hydra:projects-info sbt-output)))
+  (let ((output-cleared (replace-regexp-in-string ansi-color-regexp "" sbt-output)))
+    (setq sbt-hydra:plugins-output (concat sbt-hydra:plugins-output output-cleared))
+    (when (string-match sbt:sbt-prompt-regexp sbt-hydra:plugins-output)
+      (sbt-hydra:generate-hydras-from-projects (sbt-hydra:project-info-by-version sbt-hydra:plugins-output)))))
 
 (defun sbt-test-hydra-command:test-only (project failing-test)
-  `((sbt-test-hydra:test-only ,(format "test:testOnly %s" failing-test), project) ,(format "%s - %s" project failing-test)))
+  `((sbt-test-hydra:test-only ,(format "Test / testOnly %s" failing-test), project) ,(format "%s - %s" project failing-test)))
 
 (defun sbt-test-hydra:generate-hydra (project failing-tests)
   (let* ((sbt-commands (append
